open Lexer;;


type sml_node = SMLNode of sml_tag * sml_node list | Char of char;;


let is_blank = function
	| (' ' | '\t' | '\n' | '\r') -> true;
	| _ -> false;;

let is_color_tag = function
	| Color(_) -> true;
	| _ -> false;;
	
let is_size_tag = function
	| Size(_) -> true;
	| _ -> false;;
	
let tag_to_string = function
	| Bold -> "B"
	| Emphasis -> "EM"
	| Italic -> "I"
	| Plain -> "PL"
	| Strong -> "S"
	| Typewritter -> "TT"
	| Underline1 | Underline2 | Underline3 | Underline  -> "U"
	| Color(Red) -> "r"
	| Color(Green) -> "g"
	| Color(Blue) -> "b"
	| Color(Cyan) -> "c"
	| Color(Magenta) -> "m"
	| Color(Yellow) -> "y"
	| Color(Black) -> "k"
	| Color(White) -> "w"
	| Size(x) -> string_of_int x;;

let tag_to_size = function
	| Emphasis | Plain | Typewritter -> 2
	| _ -> 1;;

let is_bold_tag theTag = theTag = Bold;;
let is_italic_tag theTag = theTag = Italic;;
let is_strong_tag theTag = theTag = Strong;;
let is_emphasis_tag theTag = theTag = Emphasis;;
let is_typewritter_tag theTag = theTag = Typewritter;;
let is_underline_tag theTag = theTag = Underline;;
let is_underline1_tag theTag = theTag = Underline1;;
let is_underline2_tag theTag = theTag = Underline2;;
let is_underline3_tag theTag = theTag = Underline3;;

let list_count_tag theList theTag = (List.fold_left (fun sum aTag -> if aTag = theTag then sum + 1 else sum) 0 theList);;

let rec list_remove theFunction = function 
	| x::l -> if (theFunction x) then (list_remove theFunction l) else (x::(list_remove theFunction l))
	| [] -> [];;
	
let rec compact_list = 
	let compare_tags aTag bTag = match aTag, bTag with
	    | Color(_), Color(_) -> true
	    | Size(_), Size(_) -> true
	    | x,y -> x = y in
	function
	| Emphasis::l -> let newList = list_remove is_emphasis_tag l in 
				compact_list(if ((list_count_tag l Emphasis) mod 2) = 0 then Strong::newList else newList)
	| Underline1::l -> let newList = compact_list(list_remove is_underline_tag l) in
			   let underline_count = (list_count_tag l Underline) in
			    if underline_count = 0 then 
				Underline1::newList
			    else if underline_count = 1 then 
				Underline1::Underline2::newList 
			    else 
				Underline1::Underline2::Underline3::newList
	| Plain::l -> compact_list(list_remove (fun aTag -> (is_color_tag aTag || is_size_tag aTag) = false) l)
	| x::l -> x::(compact_list (list_remove (fun a -> compare_tags a x) l))
	| [] -> [];;
	
let space_compact_list theList = let newList = (list_remove (fun theTag -> (is_bold_tag theTag) || (is_italic_tag theTag) || (is_strong_tag theTag) || (is_emphasis_tag theTag)) theList) in
				if (list_count_tag newList Underline) = 0 then
				   (list_remove is_color_tag newList)
				else
				    newList;;
				    
let rec sml_node_list_to_size theList = 
    let sml_node_to_size = function
	| SMLNode(aTag, aNodeList) -> 5 + (2 * (tag_to_size aTag)) + (sml_node_list_to_size aNodeList)
	| Char(aChar) -> 1 in
    List.fold_left (fun sum aNode -> sum + (sml_node_to_size aNode)) 0 theList;;

let rec sml_node_list_print theList = 
	let rec node_print = function
	    | Char(aChar) -> print_char aChar
	    | SMLNode(aTag, aNodeList) ->   print_string ("<" ^ (tag_to_string aTag) ^ ">");
					    sml_node_list_print aNodeList;
					    print_string ("</" ^ (tag_to_string aTag) ^ ">") in
    List.iter node_print theList;;

let list_append_item theList theItem = List.rev (theItem::(List.rev theList));;

let rec is_blank_node = function
    | SMLNode(aTag, aNodeList) -> List.fold_left (fun flag node -> flag && (is_blank_node node)) true aNodeList
    | Char(x) -> is_blank x
    
let rec sml_node_list_combine_simblings theNodeList = 
	let combine_nodes aNode bNode tt_state = match aNode, bNode with
	    | SMLNode(aTag, aNodeList), SMLNode(bTag, bNodeList) when aTag = bTag -> [SMLNode(aTag, aNodeList @ bNodeList)]
	    | SMLNode(aTag, aNodeList), SMLNode(bTag, bNodeList) when ((is_color_tag aTag) && (is_color_tag bTag)) ||
								       ((is_size_tag aTag) && (is_size_tag bTag)) -> [SMLNode(aTag, list_append_item aNodeList bNode)]
	    | SMLNode(aTag, aNodeList), x when (is_blank_node x) && (aTag <> Underline1) && (aTag <> Underline2) && (aTag <> Underline1) && (aTag <> Typewritter)
								    -> [SMLNode(aTag, list_append_item aNodeList bNode)]
	    | x, SMLNode(aTag, aNodeList) when (is_blank_node x) && (aTag <> Underline1) && (aTag <> Underline2) && (aTag <> Underline1) && (aTag <> Typewritter)
								    -> [SMLNode(aTag, aNode::aNodeList)]
	    | Char(x), Char(y) when (is_blank x) && (is_blank y) && (tt_state = false) -> [aNode]
	    | _,_ -> [aNode; bNode] in
	let rec combine_siblings ?(accum = []) ?(tt_state = false) theList = 
	    match theList with
	    | x::y::l -> ( match (combine_nodes x y (tt_state || (match x with | SMLNode(Typewritter, _) -> true | _ -> false))) with
			   | [newNode] -> combine_siblings ~accum:accum (newNode::l)
			   | SMLNode(aTag, aNodeList)::_ -> combine_siblings ~accum:((SMLNode(aTag, (combine_siblings ~tt_state:(tt_state || (aTag = Typewritter)) aNodeList)))::accum) (y::l)
			   | a::_ -> combine_siblings ~accum:(a::accum) (y::l)
			   | [] -> []
			)
	    | [SMLNode(aTag, aNodeList)] -> List.rev ((SMLNode(aTag, (combine_siblings ~tt_state:(tt_state || (aTag = Typewritter)) aNodeList)))::accum)
	    | [x] -> List.rev(x::accum)
	    | _ -> List.rev(accum)
	in
	combine_siblings theNodeList;;

let rec sml_node_list_sort theFunction theList = 
	let rec node_sort theNode = match theNode with
	    | SMLNode(aTag, [x]) -> (match (node_sort x) with
				    | SMLNode(bTag, aNodeList) when (theFunction bTag) -> SMLNode(bTag, [SMLNode(aTag, aNodeList)]) 
				    | aNode -> SMLNode(aTag, [aNode]) )
	    | SMLNode(aTag, x) -> SMLNode(aTag, (sml_node_list_sort theFunction x))
	    | _ -> theNode in
	List.rev (List.rev_map node_sort theList);;

let rec sml_node_straight_tags = function
	| SMLNode(aTag, [x]) when ((is_color_tag aTag) || (is_size_tag aTag)) -> (sml_node_straight_tags x)
	| SMLNode(aTag, [x]) -> aTag::(sml_node_straight_tags x)
	| SMLNode(aTag, x) -> [aTag]
	| _ -> [];;

let rec list_union aList bList = match aList with
    | x::l when (List.exists (fun a -> a = x) bList) -> x::(list_union l bList)
    | x::l -> (list_union l bList)
    | [] -> [];;

exception Error1;;
exception Error2;;

let combine_with_plain firstNode lastNode middleList = 
	let rec node_list_from_tags theNode theTags = match theTags, theNode with
	    | [x], SMLNode(aTag, aNodeList) when aTag = x -> aNodeList
	    | x::l, SMLNode(aTag, [aNode]) when aTag = x -> node_list_from_tags aNode l
	    | x::l, SMLNode(aTag, [aNode]) -> node_list_from_tags (SMLNode(aTag, node_list_from_tags aNode [x])) l
	    | _, _ -> [theNode]  in
	let rec remove_tag_from_node theNode theTag = match theNode with
	    | SMLNode(aTag, [aNode]) when aTag = theTag -> aNode
	    | SMLNode(aTag, [aNode]) -> SMLNode(aTag, [remove_tag_from_node aNode theTag])
	    | _ -> raise Error1 in
	let rec add_node_list_to_node_after_tags theNode middleList theTags = match theTags, theNode with
	    | [x], SMLNode(aTag, aNodeList) when x = aTag -> SMLNode(aTag, aNodeList @ middleList)
	    | [x], SMLNode(aTag, [aNode]) -> SMLNode(x, (remove_tag_from_node aNode x)::middleList)
	    | x::l, aNode -> SMLNode(x, [add_node_list_to_node_after_tags  (remove_tag_from_node aNode x) middleList l])
	    | _, _ -> raise Error2 in 
	let union_tags = list_union (sml_node_straight_tags firstNode) (sml_node_straight_tags lastNode) in
	add_node_list_to_node_after_tags firstNode (SMLNode(Plain, middleList)::(node_list_from_tags lastNode union_tags)) union_tags;;

let rec sml_node_list_add_plain ?(accum = []) theList =
	let rec add_plain theNode ?(middleList = []) endList =
		let tags = sml_node_straight_tags theNode in
		if (List.length tags) < 2 then
		    theNode, endList
		else
		    (match endList with 
			| endNode::l -> let tag_union = (list_union tags(sml_node_straight_tags endNode)) in
				  if (List.length tag_union) < 2 then 
					(add_plain theNode ~middleList:(endNode::middleList) l)
				  else
					(combine_with_plain theNode endNode (List.rev middleList)), l
			| [] -> theNode, (List.rev middleList)
		    ) in
	match theList with
	| SMLNode(aTag, aNodeList)::l -> (match (add_plain (SMLNode(aTag, aNodeList)) l) with 
					    | SMLNode(newTag, newNodeList) as aNode, remainingList ->
					     if (List.length l) = (List.length remainingList) then
						    sml_node_list_add_plain ~accum:(SMLNode(newTag, (sml_node_list_add_plain newNodeList))::accum) l
						else
						    sml_node_list_add_plain ~accum:accum (aNode::remainingList)
					    | _,_ -> raise Error1
					)
	| x::l -> sml_node_list_add_plain ~accum:(x::accum) l
	| [] -> List.rev accum;;
	

let sml_document = ref [];;

let parsing_done = ref false;;

let print_lock = Mutex.create ();;

let buf_size = 8192
let input_buffer = Buffer.create buf_size;;

let print_document_and_exit aDocument = 
    Mutex.lock print_lock;
    let inputLength =  (Buffer.length input_buffer) in
    let optimizedLength = sml_node_list_to_size aDocument in 
    if (!parsing_done  = false) || (inputLength <= optimizedLength) then begin
	    prerr_endline "Printing out stdin";
	    Buffer.output_buffer stdout input_buffer;
    end else begin
	    Printf.fprintf stderr "Printing out optimized document\nstdin = %d chars\nstdout = %d chars\n" inputLength optimizedLength;
	    sml_node_list_print aDocument;
    end;
    prerr_endline "Done";
    exit(0);;

let seconds = ref 180;;
Arg.parse [] (fun secs -> seconds := (int_of_string secs)) "seconds";;


let watcher_function delay = 
	Thread.delay ((float_of_int delay) -. 3.0);
	prerr_endline "Time out!";
	print_document_and_exit !sml_document;;

let t = Thread.create watcher_function !seconds;;


let tag_combine () =
    let oldDocumentSize = ref (-1) in
    prerr_endline "Starting First Combine";
    while !oldDocumentSize != (sml_node_list_to_size !sml_document) do 
    
	oldDocumentSize := sml_node_list_to_size !sml_document;
	
	sml_document := sml_node_list_combine_simblings (sml_node_list_sort is_bold_tag !sml_document);
	sml_document := sml_node_list_combine_simblings (sml_node_list_sort is_italic_tag !sml_document);
	sml_document := sml_node_list_combine_simblings (sml_node_list_sort is_strong_tag !sml_document);
	sml_document := sml_node_list_combine_simblings (sml_node_list_sort is_typewritter_tag !sml_document);
	sml_document := sml_node_list_combine_simblings (sml_node_list_sort is_underline1_tag !sml_document);
	sml_document := sml_node_list_combine_simblings (sml_node_list_sort is_underline2_tag !sml_document);
	sml_document := sml_node_list_combine_simblings (sml_node_list_sort is_underline3_tag !sml_document);
	
    done;
    prerr_endline "Done First Combine";;

let color_combine () =
    let oldDocumentSize = ref (-1) in
    prerr_endline "Starting Second Combine";
    while !oldDocumentSize != (sml_node_list_to_size !sml_document) do 

	oldDocumentSize := sml_node_list_to_size !sml_document;
	
	sml_document := sml_node_list_combine_simblings (sml_node_list_sort is_color_tag !sml_document);
	sml_document := sml_node_list_combine_simblings (sml_node_list_sort is_size_tag !sml_document);
	
    done;
    prerr_endline "Done Second Combine";;

let pl_optimization () = 
    prerr_endline "Starting PL optimization";
    sml_document := sml_node_list_add_plain !sml_document;
    prerr_endline "Done PL optimization";;

try
    prerr_endline "Starting to Read Input";
	
    let buf_string = String.make buf_size ' '  in
    let chars_read = ref (-1) in
    while !chars_read <> 0 do 
	chars_read := input stdin buf_string 0 buf_size;
	Buffer.add_substring input_buffer buf_string 0 !chars_read;
    done;
    Printf.fprintf stderr "Done Reading Input %d chars read\n" (Buffer.length input_buffer);

    prerr_endline "Starting to Parse";
    let lexbuf = Lexing.from_string (Buffer.contents input_buffer) in
    let document = ref [] in
    try
	while true do 
	    let lexedChar = Lexer.sml_lexed_node lexbuf in
	    let aChar = fst lexedChar in
	    let aList = compact_list (if is_blank aChar then (space_compact_list (snd lexedChar)) else (snd lexedChar)) in
		document := (List.fold_left (fun aNode aTag -> SMLNode(aTag, [aNode])) (Char(aChar)) aList)::!document;
	done
    with EOF -> sml_document := List.rev (!document);
    prerr_endline "Done Parsing";
    parsing_done := true;
    
    tag_combine ();

    pl_optimization ();
    
    color_combine ();

with _ -> ();;

print_document_and_exit !sml_document;